
module Main where

-- friends
import Random.Array

import qualified Solver.Vector          as V
import qualified Solver.Accelerate      as A
import qualified Solver.CUBLAS          as B

import Data.Array.Accelerate            as A
import Data.Array.Accelerate.IO         as A
import Data.Array.Accelerate.CUDA       as A


-- standard library
import Criterion.Main
import System.Random.MWC
import Foreign.CUDA                     as CUDA
import Foreign.CUDA.BLAS                as BLAS
import qualified Data.Vector.Storable   as V


doTest :: Handle -> Int -> IO ()
doTest hdl x = do
  xs_arr        <- randomArrayIO (const uniform) (Z :. n)
  ys_arr        <- randomArrayIO (const uniform) (Z :. n)

  let ((),xs_vec) = toVectors xs_arr
      ((),ys_vec) = toVectors ys_arr

  xs_dev        <- V.unsafeWith xs_vec copyToDevice
  ys_dev        <- V.unsafeWith ys_vec copyToDevice

  -- Run the benchmarks
  defaultMain
    [ bgroup ("n=" ++ shows x "M")
--        [ bench "vector"        $ nf   (V.dotp xs_vec) ys_vec
        [ bench "cublas"        $ B.dotp hdl n xs_dev ys_dev
        , bench "accelerate"    $ whnf (A.run1 (A.dotp (A.use xs_arr))) ys_arr
        ]
    ]

  CUDA.free xs_dev
  CUDA.free ys_dev

  putStrLn "\n"
  where
    n = x * 1000000   -- x million

    copyToDevice p = do
      d <- CUDA.mallocArray n
      CUDA.pokeArray n p d
      return d



main :: IO ()
main = do
  -- Initialise CUBLAS environment
  hdl <- BLAS.create

  -- Setup and run the benchmark suite. This reinitialises the environment every
  -- time, but means we don't keep so much live memory around.
  --
  mapM_ (doTest hdl) [2,4..20]

  -- Shutdown CUBLAS and release device memory
  BLAS.destroy hdl

